from enum import Enum


class Square(Enum):
    EMPTY = 0
    FIRST = 1
    SECOND = 2

    def __invert__(self):
        return Square.EMPTY if self == Square.EMPTY else Square.SECOND if self == Square.FIRST else Square.FIRST


class Player(Enum):
    FIRST = 1
    SECOND = 2

    def __invert__(self):
        return Player.SECOND if self == Player.FIRST else Player.FIRST



from typing import Union

from .board import Board
from .enums import Square, Player


class ImpossibleMoveError(Exception):
    # when raised, board should be safe
    pass


class Game:
    winner: Player

    def __init__(self, height: int, width: int):
        self._board = Board(height, width)
        self._used = {self._board.hash()}
        self.current_move = Player.FIRST
        self.is_over = False

    @property
    def height(self) -> int:
        return self._board.height

    @property
    def width(self) -> int:
        return self._board.width

    def is_possible(self, x: int, y: int, player: Player, message: bool = False) -> Union[bool, str]:
        if self._board[x, y] == Square((~player).value):
            return 'foreign stone in the square' if message else False
        self._board[x, y] = Square(player.value) if self._board[x, y] == Square.EMPTY else Square.EMPTY
        res = self._board.hash() not in self._used
        self._board.rollback()
        if message:
            return '' if res else 'situation used before'
        return res

    def possible_moves(self, player: Player) -> set[tuple[int, int]]:
        return {(i, j) for j in range(self.width) for i in range(self.height) if self.is_possible(i, j, player)}

    def move(self, x: int, y: int):
        if self.is_over:
            raise ImpossibleMoveError('game over')
        if message := self.is_possible(x, y, self.current_move, True):
            raise ImpossibleMoveError(message)
        self._board[x, y] = Square(self.current_move.value) if self._board[x, y] == Square.EMPTY else Square.EMPTY
        self._used.add(self._board.hash())
        if not self.possible_moves(~self.current_move):
            self.is_over = True
            self.winner = self.current_move
        self.current_move = ~self.current_move

    def __getitem__(self, item):
        return self._board[item]



from .enums import Enum, Square


class RollbackError(Exception):
    pass


class Board:
    def __init__(self, height: int, width: int):
        if height <= 0 or width <= 0:
            raise ValueError('height or width not natural')
        self._board = [[Square.EMPTY] * width for _ in range(height)]
        self._last_key = (0, 0)
        self._last_square = Square.EMPTY

    @property
    def height(self) -> int:
        return len(self._board)

    @property
    def width(self) -> int:
        return len(self._board[0])

    def __getitem__(self, item: tuple[int, int]) -> Square:
        if len(item) != 2:
            raise IndexError(f'expected size of key 2, {len(item)} found')
        return self._board[item[0]][item[1]]

    def __setitem__(self, key: tuple[int, int], value):
        if len(key) != 2:
            raise IndexError(f'expected size of key 2, {len(key)} found')
        self._last_key = key
        self._last_square = self._board[key[0]][key[1]]
        if isinstance(value, Enum):
            value = value.value
        self._board[key[0]][key[1]] = Square(value)

    def rollback(self):
        if self[self._last_key] == self._last_square:
            raise RollbackError('nothing to rollback')
        self[self._last_key] = self._last_square

    def __hash__(self) -> int:
        res = 0
        for row in self._board:
            for square in row:
                res *= 5  # len(Square) + 2
                res += square.value + 1
            res += 4  # len(Square) + 1 means a newline
        return res

    @property
    def _transforms(self):
        def transforms(board_):
            def reflects(board):
                yield board
                yield board[::-1]
                yield [row[::-1] for row in board]
                yield [row[::-1] for row in board[::-1]]

            for form in reflects(board_):
                yield form
            if self.height == self.width:
                for form in reflects([[board_[j][i] for j in range(self.width)] for i in range(self.height)]):
                    yield form

        for transform in transforms(self._board):
            yield transform
        for transform in transforms([[~square for square in row] for row in self._board]):
            yield transform

    def hash(self) -> int:  # min hash of all transformations
        temp = Board(self.height, self.width)
        res = hash(self)
        for board in self._transforms:
            temp._board = board
            res = min(res, hash(temp))
        return res



from podch import Game
from podch.game import ImpossibleMoveError

if __name__ == '__main__':
    while True:
        try:
            height, width = map(int, input('Board heigth and width (e.g. 2 3): ').split())
            game = Game(height, width)
        except TypeError:
            print('Size does not match format')
        except ValueError:
            print('Size does not match format')
        else:
            break

    while not game.is_over:
        for i in range(height):
            print(''.join(str(game[i, j].value) for j in range(width)))
        while True:
            try:
                game.move(*map(int, input(f"{game.current_move}'s move (e.g. 0 1): ").split()))
            except ImpossibleMoveError as e:
                print('Impossible move: ' + e.args[0])
            except IndexError:
                print('Coordinates out of the board')
            except TypeError:
                print('Move does not match format')
            except ValueError:
                print('Move does not match format')
            else:
                break
    print(f'Game over: {game.winner} won!')



import os

import setuptools

with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'README.md')) as f:
    long_description = f.read()

with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'requirements.txt')) as f:
    install_requires = f.read().splitlines()

setuptools.setup(
    name='podch',
    version='1.0.1.2',
    author='Aleksandr Sup',
    author_email='sup@tmat.me',
    description='Abstract board game',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/SashaVonSup/podch',
    classifiers=[
        'Programming Language :: Python :: 3.9',
        'License :: Free for non-commercial use',
        'Operating System :: OS Independent',
        'Topic :: Games/Entertainment :: Board Games',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    packages=setuptools.find_packages(),
    data_files=[(os.path.dirname(os.path.realpath(__file__)), ['requirements.txt'])],
    install_requires=install_requires,
    python_requires='>=3.9',
)
